/* * * * * * * * * * * * * * * * * * * * * * * * * * * 
 *                                       L. Jezequel *
 *                                                   *
 * Rudimentary implementation of a sat-based planner *
 *                                                   *
 *                                                   *
 * Last update: 2016-03-04                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * */

import scala.io.Source
import scala.collection.mutable.Set
import scala.collection.mutable.{LinkedList => List}
import scala.collection.mutable.Map


// Main object, for running the planner  
object Planner {

  def main(args: Array[String]) {

    val fr: FileReader = new FileReader(args(0))
    val pp: PlanningProblem = fr.read

    //println(pp)
  
    val ppt: PlanningProblemTransformer = new PlanningProblemTransformer(pp)
    val sp: SatProblem = ppt.toSat(args(1).toInt)

    //println(sp.nv + ", " + sp.nc)
    println(sp)
    
  }

}

// Reading input files
// The first line is the initial set of positive litterals
// (any variable not appearing here is supposed to be associated to a negative litteral)
//   of the form "a b c ..."
// The second line is the goal set of litterals
//   of the form "[-]a [-]b [-]c ..."
// Any other line is an operator (variables are implicitely declared)
//   of the form "name : [-]a [-]b [-]c ... : [-]a [-]b [-]c ..." (pre : post)
class FileReader(_filename: String) {

  def read: PlanningProblem = {

    var c: Int = 0
    var nv: Int = 0 // number of variables found
    var no: Int = 0 // number of operators found
    val vars: Map[String, Int] = Map.empty // variables numbering
    val is: Set[Int] = Set.empty // initial state
    val gs: Set[Int] = Set.empty // goal state
    val ops: Map[String, (Int, Set[Int], Set[Int])] = Map.empty // operators
  
    for (line <- Source.fromFile(this._filename).getLines()) {

      if (c == 0) {

	val sline: Array[String] = line.split(' ')

	for (litteral <- sline) {

	  val varname = litteral

	  if (!vars.contains(varname)) {
	    nv = nv + 1
	    vars += (varname -> nv)
	  }

	  val toadd = vars.getOrElse(varname, 0)

	  is += toadd
  
	}
  
      } else if (c == 1) {

	val sline: Array[String] = line.split(' ')

	for (litteral <- sline) {

	  val varname =
	    if (litteral(0) == '-') litteral.substring(1)
	    else litteral

	  if (!vars.contains(varname)) {
	    nv = nv + 1
	    vars += (varname -> nv)
	  }

	  val toadd =
	    if (litteral(0) == '-') - vars.getOrElse(varname, 0)
	    else vars.getOrElse(varname, 0)

	  gs += toadd
  
	}

      } else {

	val sline: Array[String] = line.split(" : ")
	val preline: Array[String] = sline(1).split(' ')
	val postline: Array[String] = sline(2).split(' ')

	val name: String = sline(0)
	val pre: Set[Int] = Set.empty
	val post: Set[Int] = Set.empty
  
	for (litteral <- preline) {

	  val varname =
	    if (litteral(0) == '-') litteral.substring(1)
	    else litteral

	  if (!vars.contains(varname)) {
	    nv = nv + 1
	    vars += (varname -> nv)
	  }

	  val toadd =
	    if (litteral(0) == '-') - vars.getOrElse(varname, 0)
	    else vars.getOrElse(varname, 0)

	  pre += toadd
  
	}

	for (litteral <- postline) {

	  val varname =
	    if (litteral(0) == '-') litteral.substring(1)
	    else litteral

	  if (!vars.contains(varname)) {
	    nv = nv + 1
	    vars += (varname -> nv)
	  }

	  val toadd =
	    if (litteral(0) == '-') - vars.getOrElse(varname, 0)
	    else vars.getOrElse(varname, 0)

	  post += toadd
  
	}

	no += 1
	ops += (name -> (no, pre, post))

      }
      
      c = c + 1
      
    }

    return new PlanningProblem(nv, no, is, gs, ops)
  
  }

}


// Transforming planning problems to sat problems
class PlanningProblemTransformer(_ppb: PlanningProblem) {

  def litteralAtStep(l: Int, s: Int, nv: Int, no: Int): Int = {

    if (l < 0) -(-l + ((nv + no) * s))
    else l + ((nv + no) * s)
  
  }

  def toSat(steps: Int): SatProblem = {

    var nv: Int = 0 // number of variables
    var nc: Int = 0 // number of clauses
    val clauses: Set[Set[Int]] = Set.empty // set of clauses

    // clauses for initial state
    val is: Set[Int] = this._ppb.is
    for (i <- 1 to this._ppb.nv) {
      
      if (is.contains(i)) clauses += (Set.empty + i)
      else clauses += (Set.empty + (-i))
  
    }
    nv += this._ppb.nv
    nc += this._ppb.nv

    // clauses for goal states
    val gs: Set[Int] = this._ppb.gs
    for (l <- gs) {

      val toadd: Int = litteralAtStep(l, steps, this._ppb.nv, this._ppb.no)
      clauses += (Set.empty + toadd)
      nc += 1
  
    }

    // clauses for each time step
    for (step <- 0 until steps) {

      // clauses for operators
      for ((num, pre, post) <- this._ppb.ops.values) {

	for (l <- pre) {

	  val oclause: Set[Int] = Set.empty
	  oclause += litteralAtStep(-(num + this._ppb.nv), step, this._ppb.nv, this._ppb.no)
	  oclause += litteralAtStep(l, step, this._ppb.nv, this._ppb.no)
	  nc += 1
	  clauses += oclause
	  
	}

	for (l <- post) {

	  val oclause: Set[Int] = Set.empty
	  oclause += litteralAtStep(-(num + this._ppb.nv), step, this._ppb.nv, this._ppb.no)
	  oclause += litteralAtStep(l, step + 1, this._ppb.nv, this._ppb.no)
	  nc += 1
	  clauses += oclause
  
	}

	// enforces to do only one action per time step
	// could be modified to allow more relaxed semantics
	for ((numb, _, _) <- this._ppb.ops.values.filter(_._1 > num)) {

	  val oclause: Set[Int] = Set.empty
	  oclause += litteralAtStep(-(num + this._ppb.nv), step, this._ppb.nv, this._ppb.no)
	  oclause += litteralAtStep(-(numb + this._ppb.nv), step, this._ppb.nv, this._ppb.no)
	  nc += 1
	  clauses += oclause
  
	}
  
      }

      // clauses for variables
      for (v <- 1 to this._ppb.nv) {

	val vpclause: Set[Int] = Set.empty
	vpclause += litteralAtStep(-v, step + 1, this._ppb.nv, this._ppb.no)
	vpclause += litteralAtStep(v, step, this._ppb.nv, this._ppb.no)

	val vnclause: Set[Int] = Set.empty
	vnclause += litteralAtStep(v, step + 1, this._ppb.nv, this._ppb.no)
	vnclause += litteralAtStep(-v, step, this._ppb.nv, this._ppb.no)
  
	for ((num, _, post) <- this._ppb.ops.values) {

	  if (post.contains(v)) {

	    vpclause += litteralAtStep(num + this._ppb.nv, step, this._ppb.nv, this._ppb.no)
  
	  } else if (post.contains(-v)) {

	    vnclause += litteralAtStep(num + this._ppb.nv, step, this._ppb.nv, this._ppb.no)
  
	  }
  
	}

	clauses += vpclause
	clauses += vnclause
	nc += 2
  
      }
      
      nv += this._ppb.nv
      nv += this._ppb.no

    }
  
    return new SatProblem(nv, nc, clauses)
  
  }
  
}
  
  
// Storing planning problems in strips like form
// Any litteral is an integer, the sign states if it is positive or negative
// Litterals have a range from -nv to nv (excluding 0, so variables are in range 1..nv)
// Initial state is a set of litterals (is)
// Goal state is a set of litterals (gs)
// Operators are cuples of sets of litterals, put in a list (ops)
class PlanningProblem(_nv: Int, _no: Int, _is: Set[Int], _gs: Set[Int], _ops: Map[String, (Int, Set[Int], Set[Int])]) {

  override def toString: String = {

    " Number of variables: " + this._nv + "\n" +
    " Initial state: " + this._is + "\n" +
    " Goal state: " + this._gs + "\n" +
    " Numbe of operators: " + this._no + "\n" +
    " Operators: " + this._ops
  
  }

  def nv = this._nv
  def no = this._no
  def is = this._is
  def gs = this._gs
  def ops = this._ops
  
}


// Storing sat problems
class SatProblem(_nv: Int, _nc: Int, _clauses: Set[Set[Int]]) {

  def nc = this._nc
  def nv = this._nv
  
  override def toString: String = {
  
    val filinfo: String = "c this file has been auto-generated from a planning problem\n"
    val pbinfo: String = "p cnf " + this._nv + " " + this._nc
    var cls: String = ""

    for (c <- this._clauses) {

      cls += "\n"
  
      for (l <- c) {

	cls += l
	cls += " "
	
      }

      cls += "0"
  
    }

    filinfo + pbinfo + cls
    
  }
  
}

