/* * * * * * * * * * * * * * * * * * * * * * * * * * * 
 *                                       L. Jezequel *
 *                                                   *
 * Rudimentary implementation of a sat-based         * 
 * unfolder for planning problems                    *
 *                                                   *
 *                                                   *
 * Last update: 2016-03-11                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * */

import scala.io.Source
import scala.collection.mutable.Set
//import scala.collection.mutable.{LinkedList => List}
import scala.collection.mutable.Map


// Main object, for running the unfolder  
object Unfolder {

  def main(args: Array[String]) {

    val fr: FileReader = new FileReader(args(0))
    val pp: PlanningProblem = fr.read

    //println(pp)
  
    val ppt: PlanningProblemTransformer = new PlanningProblemTransformer(pp)
    val sp: SatProblem = ppt.toSat(args(1).toInt)

    //println(sp.nv + ", " + sp.nc)
    println(sp)
    
  }

}

// Reading input files
// The first line is the initial set of positive litterals
// (any variable not appearing here is supposed to be associated to a negative litteral)
//   of the form "a b c ..."
// The second line is the goal set of litterals
//   of the form "[-]a [-]b [-]c ..."
// Any other line is an operator (variables are implicitely declared)
//   of the form "name : [-]a [-]b [-]c ... : [-]a [-]b [-]c ..." (pre : post)
class FileReader(_filename: String) {

  def read: PlanningProblem = {

    var c: Int = 0
    var nv: Int = 0 // number of variables found
    var no: Int = 0 // number of operators found
    val vars: Map[String, Int] = Map.empty // variables numbering
    val is: Set[Int] = Set.empty // initial state
    val gs: Set[Int] = Set.empty // goal state
    val ops: Map[String, (Int, Set[Int], Set[Int])] = Map.empty // operators
  
    for (line <- Source.fromFile(this._filename).getLines()) {

      if (c == 0) {

	val sline: Array[String] = line.split(' ')

	for (litteral <- sline) {

	  val varname = litteral

	  if (!vars.contains(varname)) {
	    nv = nv + 1
	    vars += (varname -> nv)
	  }

	  val toadd = vars.getOrElse(varname, 0)

	  is += toadd
  
	}
  
      } else if (c == 1) {

	val sline: Array[String] = line.split(' ')

	for (litteral <- sline) {

	  val varname =
	    if (litteral(0) == '-') litteral.substring(1)
	    else litteral

	  if (!vars.contains(varname)) {
	    nv = nv + 1
	    vars += (varname -> nv)
	  }

	  val toadd =
	    if (litteral(0) == '-') - vars.getOrElse(varname, 0)
	    else vars.getOrElse(varname, 0)

	  gs += toadd
  
	}

      } else {

	val sline: Array[String] = line.split(" : ")
	val preline: Array[String] = sline(1).split(' ')
	val postline: Array[String] = sline(2).split(' ')

	val name: String = sline(0)
	val pre: Set[Int] = Set.empty
	val post: Set[Int] = Set.empty
  
	for (litteral <- preline) {

	  val varname =
	    if (litteral(0) == '-') litteral.substring(1)
	    else litteral

	  if (!vars.contains(varname)) {
	    nv = nv + 1
	    vars += (varname -> nv)
	  }

	  val toadd =
	    if (litteral(0) == '-') - vars.getOrElse(varname, 0)
	    else vars.getOrElse(varname, 0)

	  pre += toadd
  
	}

	for (litteral <- postline) {

	  val varname =
	    if (litteral(0) == '-') litteral.substring(1)
	    else litteral

	  if (!vars.contains(varname)) {
	    nv = nv + 1
	    vars += (varname -> nv)
	  }

	  val toadd =
	    if (litteral(0) == '-') - vars.getOrElse(varname, 0)
	    else vars.getOrElse(varname, 0)

	  post += toadd
  
	}

	no += 1
	ops += (name -> (no, pre, post))

      }
      
      c = c + 1
      
    }

    return new PlanningProblem(nv, no, is, gs, ops)
  
  }

}


// Transforming unfolding of planning problems to sat problems
class PlanningProblemTransformer(_ppb: PlanningProblem) {

  // depth (1 to max depth) -> sequence of operators -> sat variable
  val ops: Map[Int, Map[List[Int], Int]] = Map.empty

  // depth (0 to max depth) -> planning variable -> sequence of operators -> sat variable
  val vars: Map[Int, Map[Int, Map[List[Int], Int]]] = Map.empty

  // add a new sat variable nv to ops
  // where o is the corresponding op
  //       seq is the sequence of operators occuring before o
  //       d is the depth
  def addToOps(o: Int, seq: List[Int], d: Int, nv: Int) = {

    val newSeq: List[Int] = o :: seq
    if (!(ops.keySet.contains(d))) ops += (d -> Map.empty)
    ops(d) += (newSeq -> nv)
  
  }

  // add a new sat variable nv to vars
  // where v is the corresponding variable
  //       seq is the sequence of operators occuring before v
  //       d is the depth
  def addToVars(v: Int, seq: List[Int], d: Int, nv: Int) = {

    if (!vars.keySet.contains(d))
      vars += (d -> Map.empty)
    
    if (!vars(d).keySet.contains(v))
      vars(d) += (v -> Map.empty)
    
    vars(d)(v) += (seq -> nv)
  
  }

  def toSat(depth: Int): SatProblem = {

    var nv: Int = 0 // number of variables
    var nc: Int = 0 // number of clauses
    val clauses: Set[Set[Int]] = Set.empty // set of clauses

    // clauses for initial state
    val depth0VarMap: Map[Int, Map[List[Int], Int]] = Map.empty
    for (i <- 1 to this._ppb.nv) {

      nv += 1
      nc += 1
      val depth0Vari: Map[List[Int], Int] = Map.empty
      depth0Vari += (List.empty -> nv)
      depth0VarMap += (i -> depth0Vari)
  
      if (this._ppb.is.contains(i)) clauses += (Set.empty + nv)
      else clauses += (Set.empty + (-nv))
  
    }
    // special stop variable
    nv += 1
    depth0VarMap += (0 -> Map.empty)
    depth0VarMap(0) += (List.empty -> nv)
    clauses += (Set.empty + (-nv))
    vars += (0 -> depth0VarMap)

    // operators and variables at depth 1 (special case with data-structures initialization)
    // loop on operators
    for ((o, pre, post) <- this._ppb.ops.values) {

      val seq: List[Int] = List.empty
      val d: Int = 1

      // if pre is verified at previous depth, o has to be added (pre(d-1) => o(d))
      val clp: Set[Int] = Set.empty
      for (p <- pre) {

	val posP: Int = if (p < 0) -p else p
	val posCl: Int = vars(d-1)(posP)(seq)
	val clel: Int = if (p < 0) posCl else -posCl
	clp += clel
	
      }
      nv += 1
      nc += 1
      val vo = nv
      clp += vo
      addToOps(o, seq, d, vo)
      clauses += clp

      // else o must not be added (not pre(d-1))
      for (p <- pre) {

	val cln: Set[Int] = Set.empty
	val posP: Int = if (p < 0) -p else p
	val posCl: Int = vars(d-1)(posP)(seq)
	val clel: Int = if (p < 0) -posCl else posCl
	cln += clel
	cln += vo
	nc += 1
	clauses += cln
	
      }

      val seqb: List[Int] = o :: seq
	  
      // loop on variables
      for (v <- 1 to this._ppb.nv) {
	    
	    
	// give a name to the new variable
	nv += 1
	val vv: Int = nv
	addToVars(v, seqb, d, vv)
	
	// if o is fired, then post has to be added (o(d) => post(d) ^ no modif of other variables)
	if (post.contains(v)) {
	  // v appears as positive litteral in post
	  
	  val cl: Set[Int] = Set.empty
	  cl += -vo
	  cl += vv
	  nc += 1
	  clauses += cl
	  
	} else if (post.contains(-v)) {
	  // v appears as negative litteral in post
	  
	  val cl: Set[Int] = Set.empty
	  cl += -vo
	  cl += -vv
	  nc += 1
	  clauses += cl
	  
	    } else {
	      // v does not appear in post, copy previous value
	      
	      val cla: Set[Int] = Set.empty
	      cla += -vo
	      cla += vv
	      cla += -vars(d-1)(v)(seq)
	      val clb: Set[Int] = Set.empty
	      clb += -vo
	      clb += -vv
	      clb += vars(d-1)(v)(seq)
	      nc += 2
	      clauses += cla
	      clauses += clb
	      
	    }
	
      }
      // if o is not fired a special variable (0) is added to state that the unfolding should stop
      nv += 1
      val vs: Int = nv
      addToVars(0, seqb, d, vs)
      val stopcla: Set[Int] = Set.empty
      stopcla += vs
      stopcla += vo
      val stopclb: Set[Int] = Set.empty
      stopclb += -vs
      stopclb += -vo
      nc += 2
      clauses += stopcla
      clauses += stopclb
      
      // if stop was true at previous depth, o should not be fired
      val clstop: Set[Int] = Set.empty
      clstop += -vars(d-1)(0)(seq)
      clstop += -vo
      nc += 1
      clauses += clstop
      
    }


  
    // loop on depths
    for (d <- 2 to depth) {

      // loop on operators
      for ((o, pre, post) <- this._ppb.ops.values) {

	for (seq <- ops(d - 1).keys) {

	  // if pre is verified at previous depth, o has to be added (pre(d-1) => o(d))
	  val clp: Set[Int] = Set.empty
	  for (p <- pre) {

	    val posP: Int = if (p < 0) -p else p
	    val posCl: Int = vars(d-1)(posP)(seq)
	    val clel: Int = if (p < 0) posCl else -posCl
	    clp += clel
	    
	  }
	  nv += 1
	  nc += 1
	  val vo = nv
	  clp += vo
	  addToOps(o, seq, d, vo)
	  clauses += clp

	  // else o must not be added (not pre(d-1))
	  for (p <- pre) {

	    val cln: Set[Int] = Set.empty
	    val posP: Int = if (p < 0) -p else p
	    val posCl: Int = vars(d-1)(posP)(seq)
	    val clel: Int = if (p < 0) -posCl else posCl
	    cln += clel
	    cln += vo
	    nc += 1
	    clauses += cln
  
	  }

	  val seqb: List[Int] = o :: seq
	  
	  // loop on variables
	  for (v <- 1 to this._ppb.nv) {
	    
	    
	    // give a name to the new variable
	    nv += 1
	    val vv: Int = nv
	    addToVars(v, seqb, d, vv)
  
	    // if o is fired, then post has to be added (o(d) => post(d) ^ no modif of other variables)
	    if (post.contains(v)) {
	      // v appears as positive litteral in post
	      
	      val cl: Set[Int] = Set.empty
	      cl += -vo
	      cl += vv
	      nc += 1
	      clauses += cl
	      
	    } else if (post.contains(-v)) {
	      // v appears as negative litteral in post

	      val cl: Set[Int] = Set.empty
	      cl += -vo
	      cl += -vv
	      nc += 1
	      clauses += cl
	      
	    } else {
	      // v does not appear in post, copy previous value

	      val cla: Set[Int] = Set.empty
	      cla += -vo
	      cla += vv
	      cla += -vars(d-1)(v)(seq)
	      val clb: Set[Int] = Set.empty
	      clb += -vo
	      clb += -vv
	      clb += vars(d-1)(v)(seq)
	      nc += 2
	      clauses += cla
	      clauses += clb
	      
	    }
	    
	  }
	  // if o is not fired a special variable (0) is added to state that the unfolding should stop
	  nv += 1
	  val vs: Int = nv
	  addToVars(0, seqb, d, vs)
	  val stopcla: Set[Int] = Set.empty
	  stopcla += vs
	  stopcla += vo
	  val stopclb: Set[Int] = Set.empty
	  stopclb += -vs
	  stopclb += -vo
	  nc += 2
	  clauses += stopcla
	  clauses += stopclb
	  
	  // if stop was true at previous depth, o should not be fired
	  val clstop: Set[Int] = Set.empty
	  clstop += -vars(d-1)(0)(seq)
	  clstop += -vo
	  nc += 1
	  clauses += clstop
	  
	}

      }
  
    }
  
    return new SatProblem(nv, nc, clauses)
  
  }
  
}
  
  
// Storing planning problems in strips like form
// Any litteral is an integer, the sign states if it is positive or negative
// Litterals have a range from -nv to nv (excluding 0, so variables are in range 1..nv)
// Initial state is a set of litterals (is)
// Goal state is a set of litterals (gs)
// Operators are cuples of sets of litterals, put in a list (ops)
class PlanningProblem(_nv: Int, _no: Int, _is: Set[Int], _gs: Set[Int], _ops: Map[String, (Int, Set[Int], Set[Int])]) {

  override def toString: String = {

    " Number of variables: " + this._nv + "\n" +
    " Initial state: " + this._is + "\n" +
    " Goal state: " + this._gs + "\n" +
    " Numbe of operators: " + this._no + "\n" +
    " Operators: " + this._ops
  
  }

  def nv = this._nv
  def no = this._no
  def is = this._is
  def gs = this._gs
  def ops = this._ops
  
}


// Storing sat problems
class SatProblem(_nv: Int, _nc: Int, _clauses: Set[Set[Int]]) {

  def nv = this._nv
  def nc = this._nc
  
  override def toString: String = {

    val filinfo: String = "c this file has been auto-generated from a planning problem\n"
    val pbinfo: String = "p cnf " + this._nv + " " + this._nc
    var cls: String = ""

    for (c <- this._clauses) {

      cls += "\n"
  
      for (l <- c) {

	cls += l
	cls += " "
	
      }

      cls += "0"
  
    }

    filinfo + pbinfo + cls
    
  }
  
}

